/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <marlin/marlin-sample-element.h>
#include <marlin/marlin-block.h>

#include <gst/bytestream/bytestream.h>

typedef struct _SinkInputChannel {
	GstPad *sinkpad;
	GstByteStream *bytestream;
} SinkInputChannel;

GstPadTemplate *sink_template = NULL;

static GstElementClass *sink_parent_class = NULL;

/* Mmap and store a complete MARLIN_BLOCK */
static void
store_block (MarlinSample *sample,
	     int channel_number,
	     float *data)
{
	MarlinChannel *channel;
	float *mmap_data;
	
	mmap_data = marlin_sample_write_and_mmap_data (sample, data,
						       NULL);
	
	if (data == NULL) {
		g_warning ("Serious error...");
		return;
	}
		
	channel = marlin_sample_get_channel (sample, channel_number);
	marlin_channel_append_data (channel, data, MARLIN_BLOCK_SIZE);
}

static void
marlin_sample_element_sink_loop (GstElement *element)
{
	MarlinSampleElementSink *sink;
	guint j;
	GstEvent *event = NULL;
	GList *inputs;
	SinkInputChannel *channel;
	guint8 *raw_in;
	guint32 waiting;
	guint32 got_bytes, wanted_bytes;
	int i;
	
	sink = MARLIN_SAMPLE_ELEMENT_SINK (element);

	do {
		inputs = sink->input_channels;
		i = 0;

/* 		g_print ("loop - %p\n", inputs); */
		while (inputs) {
			channel = inputs->data;

			g_print ("Here?\n");
			if (!GST_PAD_IS_USABLE (channel->sinkpad)) {
				g_print ("Not usable\n");
				inputs = inputs->next;
				continue;
			}

			/* Get data from the bytestream of each input channel.
			   We need to check for events before passing on the
			   data to the sample. */
			got_bytes = gst_bytestream_peek_bytes (channel->bytestream,
							       &raw_in, MARLIN_BLOCK_SIZE);
			g_print ("Got %d bytes\n", got_bytes);
			if (got_bytes < MARLIN_BLOCK_SIZE) {
				/* Check for event */
				gst_bytestream_get_status (channel->bytestream,
							   &waiting, &event);

				if (event) {
					if (GST_EVENT_TYPE (event) == GST_EVENT_EOS) {
						/* We've got an EOS from one of
						   our pads. Remove it from
						   the list */
						gst_bytestream_destroy (channel->bytestream);
						g_free (channel);

						sink->input_channels = g_list_delete_link (inputs, inputs);
						inputs = sink->input_channels;

						break;
					}
				} else {
					g_print ("Got incomplete stream. Need to copy it\n");
				}
			} else {
				/* Here's where the data gets copied. */
				gfloat *in = (float *) raw_in;
				
				store_block (sink->sample, i, in);
				gst_bytestream_flush (channel->bytestream,
						      MARLIN_BLOCK_SIZE);
			}
			
			inputs = inputs->next;
			i++;
		}

		g_print ("Yield\n");
		gst_element_yield (element);
		g_print ("Back from yield\n");
	} while (TRUE);
}

static GstPadConnectReturn
marlin_sample_element_sink_connect (GstPad *pad,
				    GstCaps *caps)
{
	MarlinSampleElementSink *sink;
	int channels, rate;

	sink = MARLIN_SAMPLE_ELEMENT_SINK (GST_PAD_PARENT (pad));

	if (!GST_CAPS_IS_FIXED (caps)) {
		return GST_PAD_CONNECT_DELAYED;
	}
	
	gst_caps_get_int (caps, "rate", &rate);
	gst_caps_get_int (caps, "channels", &channels);
	
	g_object_set (G_OBJECT (sink->sample),
		      "sample_rate", rate,
		      "channels", channels,
		      NULL);
	
	return GST_PAD_CONNECT_OK;
}

static GstPad *
request_new_pad (GstElement *element,
		 GstPadTemplate *templ,
		 const char *unused)
{
	char *name;
	MarlinSampleElementSink *sink;
	SinkInputChannel *input_channel;
	GstPad *pad;

	sink = MARLIN_SAMPLE_ELEMENT_SINK (element);

	if  (templ->direction != GST_PAD_SINK) {
		g_warning ("MarlinSampleElementSink: Request new pad this is not a SINK");
		return NULL;
	}

	g_print ("New pad requested\n");
	input_channel = g_new (SinkInputChannel, 1);
	
	name = g_strdup_printf ("sink%d", sink->numsinkpads);
	input_channel->sinkpad = gst_pad_new_from_template (templ, name);
	g_free (name);
	input_channel->bytestream = gst_bytestream_new (input_channel->sinkpad);
	
	gst_element_add_pad (element, input_channel->sinkpad);
	gst_pad_set_connect_function (input_channel->sinkpad,
				      marlin_sample_element_sink_connect);
	
	sink->input_channels = g_list_append (sink->input_channels, input_channel);
	sink->numsinkpads++;
	
	return input_channel->sinkpad;
}

static GstPadTemplate *
sink_factory (void)
{
	return gst_pad_template_new (
		"sink%d",
		GST_PAD_SINK,
		GST_PAD_REQUEST,
		gst_caps_new (
			"float_sink",
			"audio/raw",
			gst_props_new (
				"format", GST_PROPS_STRING ("float"),
				"layout", GST_PROPS_STRING ("gfloat"),
				"intercept", GST_PROPS_FLOAT (0.0),
				"slope", GST_PROPS_FLOAT (1.0),
				"rate", GST_PROPS_INT_RANGE (4000, 96000),
				"channels", GST_PROPS_INT_RANGE (1, 4),
				NULL)),
		NULL);
}

static void
class_init (MarlinSampleElementSinkClass *klass)
{
	GObjectClass *object_class;
	GstElementClass *element_class;

	object_class = G_OBJECT_CLASS (klass);
	element_class = GST_ELEMENT_CLASS (klass);

	element_class->request_new_pad = request_new_pad;

	if (sink_template == NULL) {
		sink_template = sink_factory ();
	}
	
	gst_element_class_add_pad_template (element_class, sink_template);
					    
	sink_parent_class = g_type_class_ref (GST_TYPE_ELEMENT);
}

static void
init (MarlinSampleElementSink *sink)
{
	if (sink_template == NULL) {

	}
	
	
	gst_element_set_loop_function (GST_ELEMENT (sink),
				       marlin_sample_element_sink_loop);
	sink->numsinkpads = 0;
	sink->input_channels = NULL;
}

GType
marlin_sample_element_sink_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinSampleElementSinkClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinSampleElementSink), 0,
			(GInstanceInitFunc) init
		};

		type = g_type_register_static (GST_TYPE_ELEMENT, "MarlinSampleElementSink", &info, 0);
	}

	return type;
}

MarlinSampleElementSink *
marlin_sample_element_sink_new (MarlinSample *sample)
{
	MarlinSampleElementSink *sink;

	sink = g_object_new (MARLIN_SAMPLE_ELEMENT_SINK_TYPE, NULL);
	sink->sample = sample;

	return sink;
