/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

static void
oneton_new_pad (GstElement *oneton,
		GstPad *src,
		MarlinSample *sample)
{
	GstPad *sink, *intsrc;
	GstElement *queue, *intfloat, *autobin, *sample_sink;
	char *name, *ifname, *queuename;
	
	/* Pause the pipeline */
	gst_element_set_state (sample->priv->pipeline, GST_STATE_PAUSED);

	autobin = gst_bin_get_by_name (sample->priv->pipeline, "autobin");

	/* Add the pad as a ghost pad to the autobin */
	name = gst_pad_get_name (src);
	gst_element_add_ghost_pad (autobin, src, name);

	queuename = g_strdup_printf ("queue.%s", name);
	queue = gst_element_factory_make ("queue", queuename);
	g_free (queuename);

	gst_bin_add (GST_BIN (sample->priv->pipeline), queue);
	gst_pad_connect (gst_element_get_pad (autobin, name),
			 gst_element_get_pad (queue, "sink"));
	
	ifname = g_strdup_printf ("intfloat.%s", name);
	intfloat = gst_element_factory_make ("int2float", ifname);
	g_free (ifname);
	
	gst_bin_add (GST_BIN (sample->priv->pipeline), intfloat);
	gst_pad_connect (gst_element_get_pad (queue, "src"),
			 gst_element_get_pad (intfloat, "sink"));

	intsrc = gst_element_get_request_pad (intfloat, "src%d");
	g_assert (intsrc != NULL);

	/* Request a new dec pad */
	sample_sink = sample->priv->sample_sink;
	g_print ("Requesting pad\n");
	sink = gst_element_get_request_pad (sample_sink, "sink%d");
	g_assert (sink != NULL);
	
	/* Connect the two pads */
	if (!gst_pad_connect (intsrc, sink)) {
		g_warning ("Could not connect the two pads\n");
	}

	/* Go */
	g_print ("Setting state\n");
	gst_element_set_state (sample->priv->pipeline, GST_STATE_PLAYING);
	g_print ("Crashing\n");
}

static void
found_type (GstElement *typefind,
	    GstCaps *caps,
	    MarlinSample *sample)
{
	MarlinSamplePrivate *priv;
	GstElement *raw, *autoplug, *queue;
	GstElement *autobin, *filesrc, *cache;
	GstElement *sample_sink, *oneton;
	GstPad *sink;
	GstCaps *new_caps;
	GstProps *props;
	const char *mime;

	priv = sample->priv;
	
	mime = gst_caps_get_mime (caps);
	g_print ("Found type - %s\n", mime);

	gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);

	filesrc = gst_bin_get_by_name (GST_BIN (priv->pipeline), "file_source");
	autobin = gst_bin_get_by_name (GST_BIN (priv->pipeline), "autobin");
	cache = gst_bin_get_by_name (GST_BIN (autobin), "cache");

	/* Disconnect the typefind. */
	gst_element_disconnect_pads (cache, "src", typefind, "sink");
	gst_bin_remove (GST_BIN (autobin), typefind);

	/* Create oneton */
	oneton = gst_element_factory_make ("oneton", "oneton");
	g_print ("oneton: %p\n", oneton);
	g_signal_connect (G_OBJECT (oneton), "new-pad",
			  G_CALLBACK (oneton_new_pad), sample);
	
	/* Create an autoplugger */
	autoplug = gst_autoplug_factory_make ("staticrender");
	g_assert (autoplug != NULL);

	/* Plug them all together */
	raw = gst_autoplug_to_renderers (autoplug,
					 caps,
					 oneton,
					 NULL);
	if (raw == NULL) {
		g_warning ("Could not autoplug");
		return;
	}

	gst_element_set_name (raw, "raw");
	/* Add the new element to the bin */
	gst_bin_add (GST_BIN (autobin), raw);

	/* Reset the cache */
	g_object_set (G_OBJECT (cache),
		      "reset", TRUE,
		      NULL);

	/* Connect the new element up */
	gst_element_connect_pads (cache, "src", raw, "sink");

	/* Create a MarlinSampleElementSink for the sample */
	sample_sink = marlin_sample_element_sink_new (sample);
	sample->priv->sample_sink = sample_sink;
	gst_element_set_name (sample_sink, "samplesink");
	gst_bin_add (GST_BIN (sample->priv->pipeline), sample_sink);

	/* Let's go Mr Driver... */
  	gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
}

static void
cache_empty (GstElement *element,
	     MarlinSample *sample)
{
	MarlinSamplePrivate *priv;
	GstElement *autobin, *filesrc, *cache, *raw;
	GstPad *ghost;

	priv = sample->priv;

	g_print ("Cache removed\n");
	gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);

	filesrc = gst_bin_get_by_name (GST_BIN (priv->pipeline), "file_source");
	autobin = gst_bin_get_by_name (GST_BIN (priv->pipeline), "autobin");
	cache = gst_bin_get_by_name (GST_BIN (autobin), "cache");
	raw = gst_bin_get_by_name (GST_BIN (autobin), "raw");

	gst_element_disconnect_pads (filesrc, "src", cache, "sink");
	gst_element_disconnect_pads (cache, "src", raw, "sink");
	gst_bin_remove (GST_BIN (autobin), cache);

	ghost = gst_element_get_pad (autobin, "sink");
	gst_element_remove_ghost_pad (autobin, ghost);
	gst_element_add_ghost_pad (autobin,
				   gst_element_get_pad (raw, "sink"),
				   "sink");
	
	gst_element_connect_pads (filesrc, "src", autobin, "sink");

 	gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
}

int
marlin_sample_load_file (MarlinSample *sample,
			 const char *filename,
			 GError **error)
{
	GstElement *bin, *filesrc, *typefind, *cache, *autobin;
	char *basename;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);
	g_return_val_if_fail (filename != NULL, FALSE);

	g_object_set (G_OBJECT (sample),
		      "filename", filename,
		      NULL);

	basename = g_path_get_basename (filename);
	sample->priv->mmap_file = g_strdup_printf ("%s/%s.XXXXXX",
						   tmp_dir, basename);
	g_free (basename);
	
	sample->priv->mmap_fd = mkstemp (sample->priv->mmap_file);
	if (sample->priv->mmap_fd == -1) {
		if (error == NULL) {
			g_warning ("Error opening %s", sample->priv->mmap_file);
		} else {
			/* GError */
		}
		
		return 0;
	}
	
	/* create a new bin to hold the elements */
	bin = gst_thread_new ("pipeline");
	sample->priv->pipeline = bin;
	g_print ("pipeline: %p\n", bin);
	
	filesrc = gst_element_factory_make ("filesrc", "file_source");
	g_print ("filesrc: %p\n", filesrc);
	sample->priv->filesrc = filesrc;
	g_object_set (G_OBJECT (filesrc),
		      "location", filename,
		      NULL);
	g_signal_connect (G_OBJECT (filesrc), "eos",
			  G_CALLBACK (eos_callback), sample);

	gst_bin_add (GST_BIN (bin), filesrc);

	autobin = gst_bin_new ("autobin");
	g_print ("autobin: %p\n", autobin);
	cache = gst_element_factory_make ("autoplugcache", "cache");
	g_print ("cache: %p\n", cache);
	g_signal_connect (G_OBJECT (cache), "cache-empty",
			  G_CALLBACK (cache_empty), sample);
	gst_bin_add (GST_BIN (autobin), cache);

	typefind = gst_element_factory_make ("typefind", "typefind");
	g_print ("typefind: %p\n", typefind);
	g_signal_connect (G_OBJECT (typefind), "have-type",
			  G_CALLBACK (found_type), sample);
	gst_bin_add (GST_BIN (autobin), typefind);

	gst_element_connect_pads (cache, "src", typefind, "sink");
	gst_element_add_ghost_pad (autobin,
				   gst_element_get_pad (cache, "sink"),
				   "sink");

	gst_bin_add (GST_BIN (bin), autobin);
	gst_element_connect_pads (filesrc, "src", autobin, "sink");

	marlin_sample_start_load (sample);
	/* Start playing */
	gst_element_set_state (GST_ELEMENT (bin), GST_STATE_PLAYING);

	return 0;
}
